/****************************
		'resource.c'
****************************/

#include "std.h"
#include "file.h"
#include "resource.h"
#include "error.h"
#include "mem.h"

FHEAD fhead;
HEAD *rsctop;

/*** Open resource ***/

HEAD *openrsc(keep,farkeep)
word keep;
word farkeep;
{
	byte far *fp;
	HEAD *rh;

	rh=NIL;
	fp=NIL;
	rh=malloc(keep+HEADSZ);
	if(!rh) goto skip;
	if(farkeep) {
		fp=farmalloc((ulong)farkeep);
		if(!fp) {
			free(rh);
			rh=NIL;
			goto skip;
		}
	}
	memset(rh,0,keep+HEADSZ);
	rh->keep=keep;
	if(rh->farp=fp) {
		rh->farkeep=farkeep;
		farmemset(fp,0,farkeep);
	}
	addrsc(rh);
	rh->status=RST_NEW;
skip:
	if(!rh) nomem();
	return(rh);	
}

/*** Close resource ***/

closersc(rh)
HEAD *rh;
{
	HEAD *rh1;

	rh1=rsctop;
	if(rh == rh1) {
		rsctop=NIL;
	} else {
		while(rh1 && rh1->link != rh) rh1=rh1->link;
		if(rh1) {
			rh1->link=rh->link;
		}
	}
	if(rh->farp) {
		farfree(rh->farp);
	}
	free(rh);
}

/*** Load resource ***/

HEAD *loadrsc(name)
{
	int  fh;
	word size;
	byte *p;
	byte far *fp;
	HEAD *rh;

	rh=NIL;
	if((fh=_open(name,O_RDONLY)) >= 0) {
		size=_read(fh,&fhead,FHEADSZ);
		if(size == FHEADSZ) {
			 if(fhead.id == HEADID) {
				p=malloc(fhead.keep+HEADSZ);
				if(!p) { nomem(); goto skip; }
				memset(p,0,fhead.keep+HEADSZ);
				memcpy(p+RHEADSZ,&fhead,FHEADSZ);
				size=_read(fh,p+HEADSZ,fhead.size);
				if(size != fhead.size) goto skip;
				if(fhead.farkeep) {
					fp=farmalloc((ulong)fhead.farkeep);
					if(!fp) { nomem(); goto skip; }
					farmemset(fp,0,fhead.farkeep);
					size=farread(fh,fp,fhead.farsize);
					if(size != fhead.farsize) goto skip;
				} else {
					fp=NIL;
				}
				rh=(HEAD*)p;
				rh->farp=fp;
				strcpy(rh->name,name);
				strlwr(rh->name);
				rh->status=0;
				addrsc(rh);
			} else {
				errno=EINVFMT;
			}
		}
skip:
		_close(fh);
		error();
	}
	errno=0;
	cld();
	return(rh);	
}

/*** Save resource ***/

flag saversc(rh,pack)
HEAD *rh;
flag pack;
{
	flag done;
	int  fh;
	word size;

	done=FALSE;
	if((fh=openw(rh->name)) >= 0) {
		rh->size=pack ? membody((byte*)rh+HEADSZ,rh->keep) : rh->keep;
		if(rh->farp) {
			rh->farsize=pack ? farmembody(rh->farp,rh->farkeep) : rh->farkeep;
		}
		rh->id=HEADID;
		rh->total=FHEADSZ+rh->size+rh->farsize;
		size=_write(fh,(byte*)rh+RHEADSZ,rh->size+FHEADSZ);
		if(size == rh->size+FHEADSZ) {
			if(rh->farsize) {
				size=farwrite(fh,rh->farp,rh->farsize);
				if(size == rh->farsize) {
					done=TRUE;
				}
			} else {
				done=TRUE;
			}
		}
		_write(fh,rh,0);
		close(fh);
	}
	if(done) rh->status&=~RST_MDF;
	else {
		errno=ENOSPC;
		error();
	}
	error();
	cld();
	return(done);
}

/*** Add resource ***/

addrsc(newrh)
HEAD *newrh;
{
	HEAD *rh;

	rh=rsctop;
	if(!rh) {
		rsctop=newrh;
	} else {
		while(rh->link) rh=rh->link;
		rh->link=newrh;
	}
	newrh->link=NIL;
}

/*** Allocate resource ***/

HEAD *allocrsc(rh,keep,farkeep)
HEAD *rh;
word keep;
word farkeep;
{
	byte far *fp;
	HEAD *newrh;
	HEAD *rh1;

	if(keep != rh->keep) {
		newrh=(HEAD*)realloc(rh,keep+HEADSZ);
		if(!newrh) { rh=NIL; goto skip; }
		if(rh != newrh) {
			rh1=rsctop;
			if(rh == rh1) {
				rsctop=newrh;
			} else {
				while(rh1 && rh1->link != rh) rh1=rh1->link;
				if(rh1) {
					rh1->link=newrh;
				}
			}
		}
		rh=newrh;
		if(keep > rh->keep) {
			memset((byte*)rh+rh->keep+HEADSZ,0,keep-rh->keep);
		}
		rh->keep=keep;
	}
	if(rh->farp && farkeep != rh->farkeep) {
		fp=farrealloc(rh->farp,(ulong)farkeep);
		if(!fp) { rh=NIL; goto skip; }
		if(fp != rh->farp) {
			rh->farp=fp;
		}
		if(farkeep > rh->farkeep) {
			farmemset(fp+rh->farkeep,0,farkeep-rh->farkeep);
		}
		rh->farkeep=farkeep;
	}
skip:
	if(!rh) nomem();
	return(rh);
}

/*** Clear resource (Fill by 0) ***/

clearrsc(rh)
HEAD *rh;
{
	memset((byte*)rh+HEADSZ,0,rh->keep);
	farmemset(rh->farp,0,rh->farkeep);
	rh->status|=RST_NEW;
	rh->status&=~RST_MDF;
}

/*** Scan head ptr from resource No. ***/

HEAD *headrsc(n)
int  n;
{
	int  i;
	HEAD *rh;

	rh=rsctop;
	if(rh) {
		for_to(i,n) {
			rh=rh->link;
			if(!rh) break;
		}
	}
	return(rh);
}

/*** out of memory ***/

static nomem()
{
	errno=ENOMEM;
	error();
}

/*** Search RSC name ***/

HEAD *searchrsc(name)
char *name;
{
	HEAD *rh;
	
	rh=rsctop;
	while(rh) {
		if(!strcmp(rh->name,name)) break;
		rh=rh->link;
	}
	return(rh);
}
