/************************
	'MET' Edit Module
************************/

#include "std.h"
#include "graf.h"
#include "met.h"

#define REF_P	1
#define REF_E	2
#define TNUM	19
#define NOTESZ	28
#define LFTX	8

#define IC_REST	1
#define IC_INS	2
#define IC_BLK	3
#define IC_CUT	4
#define	IC_REP	5
#define	IC_REP1	6
#define	IC_REP2	7
#define IC_INOTE 22
#define IC_IREST 29

#define ON_REP	-3
#define ON_LFT	-2
#define ON_GAP	-1
#define ON_PORT	4
#define PORTDX	10

#define R_NOP	0
#define R_MESR	1
#define R_SCOR	2

int ticks[]={
	0,192,96,48,24,12,6,3,
	0,288,144,72,36,18,9,0,
	0,336,168,84,42,21,0,0,
	0,128,64,32,16,8,4,2
};
int  tickc[TNUM]={ 0,0,6,9,12,18,21,24,36,42,48,72,84,96,144,168,192,288,336 };
extern int barsz[];
extern int barskp[];

static char keyflg[4]={ 0,1,-1,0 };
static byte keybit[9]={ 0xC0,0,0xC0,0x40,0x40,0,0,0x80,0x80 };
static char keyshp[12]={ 2,0,1,1,1,1,2,2,2,0,0,0 };

/*** Set x pos from ptr ***/

int  setxpos(p,q)
char *p;
char *q;
{
	return(sumtox(sumtick(p,q)));
}

/*** Sum Tick ***/

int  sumtick(p,q)
char *p;
char *q;
{
	int  i;
	int  d0;
	int  tick;
	int  tsum;
	int  trep;
	int  repc;

	tsum=trep=0;
	while(p != q) {
		if(!(d0=*p)) break;
		p+=2;
		if(d0 < 0) {
			tick=settick(d0);
			tsum+=tick;
			trep+=tick;
			if(cp.chord) p+=2;
		} else {
			if(d0 == D_REPT) trep=0;
			else if(d0 == D_REPT9) {
				repc=*(p-1);
				for_to(i,repc) tsum+=trep;
			}
		}
	}
	return(tsum);
}

/*** Sum to x pos ***/

int  sumtox(tsum)
int  tsum;
{
	return(muldiv(mesrsx,tsum,cp.tmesr));
}

/*** Set Initial Sum ***/

int setisum(n)
int  n;
{
	int  tsum;

	if(!n && mb.beat) tsum=mb.beat*cp.tbeat;
	else tsum=0;
	return(tsum);
}

/*** Set Tick count ***/

int  settick(ldat)
int  ldat;
{
	int  tick;

	tick=ticks[ldat & 0x1F];
	if(!tick) tick=cp.tmesr;
	return(tick);
}

/*** Increment ptr ***/

char *incptr(p)
char *p;
{
	while(*p > 0) p+=2;
	return(p);
}

/*** X to Tick ***/

int  xtotick(x,tblk,tsum,imc)
int  x;
int  tblk;
int  tsum;
flag imc;
{
	int tc;
	int tcm;
	int  i;
	int  j;

	if(x < 0) tc=0;
	else {
		tc=muldiv(x,cp.tmesr,mesrsx);
		if(tc > tblk) tc=tblk;
	}
	if(imc) {
		tcm=cp.tbeat;
		while(tcm < tc+tsum) tcm+=cp.tbeat;
		tcm-=tsum;
	} else tcm=tblk;
	for_to(i,TNUM) {
		if(tickc[i] >= tc) break;
	}
	if((i == TNUM) || (tc != tickc[i])) i--;
	loop {
		tc=tcm-tickc[i];
		for_to(j,TNUM) {
			if(tickc[j] == tc) break;
		}
		if(j < TNUM) break;
		i--;
		if(i < 0) {
			tc=-1;
			goto end_xto;
		}
	}
	tc=tickc[i];
end_xto:
	return(tc);
}

/*** Tail Tick ***/

tailtick(tsum,tpre)
int  tsum;
int  tpre;
{
	int  i;
	int  tc;

	tsum+=tpre;
	if(tsum >= cp.tmesr) i=0;
	else {
		tc=cp.tbeat;
		while(tc <= tsum) tc+=cp.tbeat;
		tc=tc-tsum;
		for_to(i,TNUM) {
			if(tickc[i] >= tc) break;
		}
		if(tickc[i] != tc) i--;
	}
	return(tickc[i]);
}

/*** Y to Key ***/

ytokey(y)
int  y;
{
	int  key;
	int  oct;
	int  tone;

	key=cp.key_h-(y-cp.vtmgn+6)/(GRID/2);
	if(key > cp.key_h+cp.legmax) key=cp.key_h+cp.legmax;
	if(key < cp.key_l-cp.legmax) key=cp.key_l-cp.legmax;
	oct=key/7;
	tone=key%7;
	return((oct<<3)+tone);
}

/*** Tick to Lengh ***/

tolen(tick)
int  tick;
{
	int  i;

	for_to(i,32) {
		if(ticks[i] == tick) break;
	}
	return(0x80+i);
}

/*** On Note(rest) ***/
/* mode -3:on Repeat
		-2:in left area
		-1:in Gap
		 0:in note (but not on note)
		 1:on note or rest
		 2,3:on note of chord
		 4:on portament
*/
static char *prep,*edtp;
static int  octa;
static int  port;

int  onnote(cx,cy,p,n)
int  cx,cy;
char *p;
int  n;
{
	int  i;
	int  d0;
	int  d1;
	int  x;
	int  key;
	int  ckey;
	int  tick;
	int  tsum;
	int  mode;
	char *q;
	flag tri;
	flag trip;

	ckey=cp.key_h-(cy-cp.vtmgn)/(GRID/2)-2;
	octa=cp.octa;
	tsum=setisum(n);
	x=sumtox(tsum);
	prep=NIL;
	port=tri=trip=FALSE;
	if(cx < x-2) { mode=ON_LFT; edtp=p; goto end_on; }
	mode=ON_GAP;
	cp.repm=cp.trep=0;
	do {
		while(cp.repm) {
			if(cx >= x && cx < x+12) {
				mode=ON_REP;
				edtp=p;
				goto end_on;
			}
			tsum+=cp.trep;
			x=sumtox(tsum);
			if(cx < x) {
				mode=ON_LFT;
				edtp=p;
				goto end_on;
			}
			if(!--cp.repc) {
				cp.repm=cp.trep=0;
				prep=NIL;
			}
		}
		d0=*p++;
		d1=*p++;
		if(d0 < 0) {
			tri=((d0 & 0x18) == 0x18);
			if(cx >= x-2 && cx < x+10) {
				if(d1 == RESTC) mode=1;
				else {
					mode=0;
					q=p-1;
					for_to(i,(cp.chord ? 3 : 1)) {
						if((d1=*q++) == RESTC) break;
						key=((d1 & 0x38)>>3)*7+(d1 & 7);
						if(key == ckey || key == ckey+1) {
							mode=i+1;
							break;
						}
					}
				}
				port=FALSE;
				break;
			}
			else if(cx < x) {
				port=tri=FALSE;
				break;
			}
			prep=p-2;
			trip=tri;
			if(cp.chord) p+=2;
			tick=settick(d0);
			tsum+=tick;
			cp.trep+=tick;
			x=sumtox(tsum);
			if(port) {
				if(cx < x && cx >= x-PORTDX) break;
			}
			port=FALSE;
			if(d1 != RESTC && d1&0xC0) {
				kref[d1&7]=keyflg[d1>>6&3];
			}
		} else {
			switch(d0) {
			case D_OCTA:
				octa=d1;
				break;
			case D_VOICE:
				cp.voice=d1;
				break;
			case D_PORT:
				port=(int)p-2;
				break;
			case D_REPT:
				cp.trep=0;
				break;
			case D_REPT9:
				if(cx < x) { tri=FALSE; break; }
				setiparm(d0,d1);
				break;
			}
		}
	} while(d0);
	edtp=p-2;
end_on:
	if(port) {
		edtp=(char*)port;
		prep=p;
		mode=ON_PORT;
	} else {
		if(!(prep) || tri || trip) prep=edtp;
	}
	return(mode);
}

/*** Do Edit ***/

do_edit()
{
	int  i;
	int  n;
	int  x,y,x1,y1;
	int  px,py,by;
	int  xi,yi,pi;
	int  pc;
	int  mode;
	int  pre_k,pre_l;
	int  edt_k,edt_k1,edt_l;
	int  pre_g,edt_g;
	int  pregp,edtgp;
	int  blkx,blksx,edtx,skpx;
	int  tsum,tpre,tpre1,tedt,tblk;
	int  ref;
	int  sharp;
	int  chords;
	int  ret;
	flag f;
	flag trip;
	flag imcomp,imc;
	flag nins;
	char *p;
	char edt_c[4];
	PART *pp;

	f=FALSE;
	x=msrec.px-SCOR_X;
	y=msrec.py-SCOR_Y;
	if(!inrect(x,y,SCOR_SX,SCOR_SY)) goto end_do;
	f=TRUE;
	if(!(msrec.evnt & MS_LDN)) {
		if(!(insblk && msrec.evnt & MS_MOV
		 && msrec.btn & MS_LON)) goto end_do;
	}
	if(!mb.part) goto end_do;
	xi=yi=pi=0;
	px=py=by=0;
	for_to(xi,mb.barlin) {
		px+=barsz[xi];
		if(x < px) break;
	}
	px-=barsz[xi];
	for_to(yi,mulstaf) {
		by+=stafssy;
		if(y < by) break;
	}
	by-=stafssy;
	py+=by;
	pp=&part[0];
	for_to(i,mb.part) {
		if(pp->disp) {
			py+=pp->stafsy;
			if(y < py) break;
			pi++;
		}
		pp++;
	}
	py-=pp->stafsy;
	n=xi+yi*mb.barlin;
	pc=i;
	loadprec(pc);
	if(!(cp.nowp=(char*)incbar(cp.dspp,n,ndsp))) goto end_do;
	scanipar(cp.nowp);
	skpx=barskp[pi*mb.barlin*mulstaf+n];
	px+=skpx;
	x-=px;
	y-=py;
	px+=SCOR_X;
	py+=SCOR_Y;
	mesrsx=barsz[xi]-skpx;

	fillmem(kref,7,0);
	if(!cp.repm) mode=onnote(x,y,cp.nowp,ndsp+n);
	else mode=ON_REP;
	imcomp=(sumtick(cp.nowp,0) < cp.tmesr);
	if(imode) {
		if(mode != ON_REP) {
			tcsr.part=i;
			tcsr.bar=ndsp+n;
			tcsr.tick=sumtick(cp.nowp,edtp);
			if(tcsr.tick >= cp.tmesr) {
				tcsr.bar++;
				tcsr.tick=0;
			}
			tcsr.sumx=sumtox(tcsr.tick+setisum(ndsp+n))-2;
			tcsr.px=px+tcsr.sumx;
			tcsr.py=py-by;
			tcsr.sy=cp.stafsy;
			drawicsr(imode);
			setimnu();
		}
		goto end_do;
	} else if(icon >= IC_CUT && icon < IC_INOTE) {
		if(icon <= IC_REP2) {
			if(icon == IC_CUT && !(mode == ON_REP)) {
				if(mode < 0) goto end_do;
				deldata(edtp);
				if(cp.chord) deldata(edtp);
				ret=R_SCOR;
			} else ret=dorept(edtp,mode,i,n+ndsp);
		} else ret=domark(edtp,mode);
		cp.repm=0;
		if(ret == R_MESR) {
			drawmesr(cp.nowp,n,TRUE);
		} else if(ret == R_SCOR) {
			drawstf1(pc);
			drawseek();
		}
		goto end_do;
	}
	if(mode <= ON_LFT) goto end_do;

	chords=0;
	trip=nins=FALSE;
	tsum=sumtick(cp.nowp,prep)+setisum(ndsp+n);
	blkx=sumtox(tsum);
	if(port) {
		blkx-=PORTDX;
		blksx=PORTDX;
		edt_k=*(edtp+1);
		tedt=1;
		tpre=0;
	} else {
		edt_g=0;
		if(mode == ON_GAP) {
			if((*prep & 0x18) == 0x18) goto end_do;
			if(icon >= IC_INOTE) {
				i=icon-IC_INOTE+1;
				if(icon >= IC_IREST) i-=IC_IREST-IC_INOTE;
				blkx=x;
				tblk=tedt=ticks[i];
				tpre=0;
				nins=TRUE;
			} else {
				imc=imcomp && (*edtp == 0);
				tblk=imc ? (cp.tmesr-tsum) : settick(*prep);
				tpre=xtotick(x-blkx,tblk,tsum,imc);
				if(imc) tedt=tailtick(tsum,tpre);
				else tedt=tblk-tpre;
			}
			edt_k=rest ? RESTC : ytokey(y);
		} else {
			imc=imcomp && (*incptr(edtp) == 0);
			tpre=(prep == edtp) ? 0 : settick(*prep);
			if((*edtp & 0x18) == 0x18) {
				trip=TRUE;
				tblk=tedt=settick(*edtp);
			} else if(imc) {
				tedt=tailtick(tsum,tpre);
				tblk=cp.tmesr-tsum;
			} else {
				tedt=settick(*edtp);
				tblk=tpre+tedt;
			}
			if(rest || !cp.chord) {
				edt_k=*(edtp+1);
				if(rest) edt_k=RESTC;
				else {
					if(!mode || edt_k == RESTC) edt_k=ytokey(y);
				}
			} else {
				p=edtp+1;
				for_to(i,3) edt_c[i]=RESTC;
				for_to(i,3) {
					if((edt_c[i]=*p++) == RESTC) break;
					chords++;
				}
				if(mode && chords) chords=mode;
				else {
					if(chords == 3) goto end_do;
					if(cp.chord == 1 && chords == 2) goto end_do;
					edt_c[chords++]=ytokey(y);
				}
				edt_k=0;
			}
			if(edt_k != RESTC) edt_g=*edtp & 0x60;
		}
		pre_g=0;
		if(prep == edtp) pre_k=RESTC;
		else {
			pre_k=*(prep+1);
			pre_g=*prep & 0x60;
		}
		blksx=sumtox(tblk);
	}
	if(blkx+blksx > mesrsx) blksx=mesrsx-blkx;
	sharp=setshp(edt_k);
	mousoff();
	setmscsr(MC_NOTE+rest,FALSE);
	drawbase(blkx+px-LFTX,py,blksx+LFTX);
	pn.lin=0xCCCC;
	pn.op=GP_XOR;
	framerct(blkx+px-LFTX,py+MSG_SY,blksx+LFTX,cp.stafsy-MSG_SY*2);
	pn.lin=0xFFFF;
	pn.op=GP_PUT;
	mouson();
	ref=REF_P+REF_E;
	pregp=edtgp=0;
	setvch();
	loop {
		edtx=blkx+sumtox(tpre);
		pre_l=tolen(tpre)|pre_g;
		edt_l=port ? 0 : tolen(tedt)|edt_g;
		if(ref & REF_P+REF_E) {
			outkey(chords ? edt_c[chords-1] : edt_k);
			mousoff();
			if(ref & REF_P && tpre) {
				pregp=drawblk(blkx+px,py,pre_k,pre_l,cp.chord ? prep+1 : 0);
			}
			if(ref & REF_E && tedt) {
				edtgp=drawblk(edtx+px,py,edt_k,edt_l,chords ? edt_c : 0);
			}
			mouson();
		}
		ref=0;
		while(!getmous(&msrec)) {}
		if(msrec.evnt&MS_LUP) break;
		x1=msrec.px-px;
		y1=msrec.py-py;
		if(!rest) {
			if(msrec.evnt & MS_RDN) {
				if(++sharp == 3) sharp=0;
			}			
			edt_k1=ytokey(y1);
			edt_k1|=keybit[sharp*3+kref[edt_k1&7]+1];
			if(chords) {
				if(edt_k1 != edt_c[chords-1]) {
					edt_c[chords-1]=edt_k1;
					ref|=REF_E;
				}
			} else {
				if(edt_k1 != edt_k) {
					edt_k=edt_k1;
					ref|=REF_E;
				}
			}
		}
		y=y1;
		if(x1 != x && !trip && !port && !nins) {
			x=x1;
			tpre1=xtotick(x-blkx,tblk,tsum,imc);
			if(tpre1 >= 0 && tpre1 != tpre) {
				if(tsum+tpre1 <= cp.tmesr) {
					tpre=tpre1;
					tedt=imc ? tailtick(tsum,tpre) : tblk-tpre;
					ref|=REF_P+REF_E;
				}
			}
		}
		if(ref & REF_E+REF_P) {
			mousoff();
			if(ref & REF_E && edtgp) { putrect(edtgp); edtgp=0; }
			if(ref & REF_P && pregp) { putrect(pregp); pregp=0; }
			mouson();
		}
	}
	if(port) {
		*(edtp+1)=edt_k;
	} else if(nins) {
		insdata(edtp,edt_l,edt_k);
		if(cp.chord) {
			insdata(edtp+2,RESTC,RESTC);
		}
	} else {
		if(mode == ON_GAP || prep == edtp) edtp=NIL;
		if(tpre) {
			if(*prep) {
				*prep++=pre_l;
				*prep++=pre_k;
				if(cp.chord) prep+=2;
			} else {
				insdata(prep,pre_l,pre_k);
				prep+=2;
				if(cp.chord) {
					insdata(prep,RESTC,RESTC);
					prep+=2;
				}
			}
		} else {
			deldata(prep);
			if(edtp) edtp-=2;
			if(cp.chord) {
				deldata(prep);
				if(edtp) edtp-=2;
			}
		}
		if(tedt) {
			if(cp.chord && chords) {
				setchord(edt_c);
				if(edtp) {
					*edtp++=edt_l;
					for_to(i,3) *edtp++=edt_c[i];
				} else {
					insdata(prep,edt_l,edt_c[0]);
					insdata(prep+2,edt_c[1],edt_c[2]);
				}
			} else {
				if(edtp) {
					*edtp++=edt_l;
					*edtp++=edt_k;
					if(cp.chord) {
						*edtp++=RESTC;
						*edtp++=RESTC;
					}
				} else {
					insdata(prep,edt_l,edt_k);
					if(cp.chord) insdata(prep+2,RESTC,RESTC);
				}
			}
		} else if(edtp) {
			deldata(edtp);
			if(cp.chord) deldata(edtp);
		}
	}	
	mousoff();
	setmscsr(moustyp,FALSE);
	if(edtgp) putrect(edtgp);
	if(pregp) putrect(pregp);
	skeyoff(vch);
	pn.lin=0xCCCC;
	pn.op=GP_XOR;
	framerct(blkx+px-LFTX,py+MSG_SY,blksx+LFTX,cp.stafsy-MSG_SY*2);
	pn.lin=0xFFFF;
	pn.op=GP_PUT;
	if(nins && !imcomp) drawstf1(pc);
	else drawmesr(cp.nowp,n,TRUE);
	if(ndsp+n == cp.barc) {
		pp->barc++;
		drawseek();
	}
	mouson();
end_do:
	return(f);
}

/*** Draw Block ***/

int  drawblk(x,y,k,l,chord)
int  x,y;
int  k,l;
int  chord;
{
	int  retp;

	retp=getrect(x-10,y,NOTESZ,cp.stafsy);
	if(chord && k != RESTC) drawcord(x,y,chord,l,TRUE,FALSE);
	else drawnote(x,y,k,l,TRUE,FALSE);
	return(retp);
}

/*** Set Chord ***/

static setchord(p)
char *p;
{
	char *q;

	q=p+1;
	if(*p==*q) {
		*q=*(q+1);
		*(q+1)=RESTC;
	} else q++;
	if(*p==*q) *q=RESTC;
	else {
		p++;
		if(p != q) {
			if(*p==*q) *q=RESTC;
		}
	}
}

/*** Output Voice ***/

static setvch()
{
	if(cp.clef == CLEF_N) vch=6;
	else vch=(cp.mode & 0x80) ? 3 : 0;
	setvoice(vch,(char*)vheadp+HEADSZ+cp.voice*VBLKSZ+16);
	setksig(cp.ksig);
}

static outkey(key)
word key;
{
	int  sharp;

	if(key != RESTC) {
		key+=octa<<3;
		sharp=setshp(key)<<6;
		key&=0x3F;
		key|=sharp;
		outsnd(key);
	}
}

static int setshp(key)
int  key;
{
	return((int)keyshp[(key>>6&3)*3+kref[key&7]+1]);
}
