/************************
	'MET' Insert Item
************************/

#include "std.h"
#include "graf.h"
#include "met.h"

#define IC_BEAMC 37
#define IC_BEAMS 38
#define IC_TEMPO 39
#define IC_OCTA	40
#define IC_VOICE 41
#define IC_GS	42
#define IC_LFO	43
#define IC_CLEF	44
#define IC_KSIG	45
#define IC_TSIG	46
#define IC_BAR	47
#define IC_BAR9	57

#define M_TEMPO 0x01
#define M_OCTA	0x02
#define M_VOICE 0x04
#define M_GS	0x08
#define M_LFO	0x10
#define M_CLEF	0x20
#define M_KSIG	0x40
#define M_TSIG	0x80

#define INSBARC	10
#define PALSZ	20
#define I_BAR	20

static DLOG dg_clef={ "Insert Clef...\n\n",180,8*25,3,Q_OK,Q_ESC,Q_DEL };
static BTNS bt_clef1={ CT_RBTN,0,0,0,0,0,2,HZ,8 };
static BTN  rb_clef1[2]={
	{ "G Clef",0,1,0 },
	{ "F Clef",0,1,0 }
};
static BTNS bt_clef2={ CT_RBTN,0,0,0,0,0,5,VT,4 };
static BTN  rb_clef2[5]={
	{ "15va up",0,1,0 },
	{ " 8va up",0,1,0 },
	{ "Normal",0,1,0 },
	{ " 8va down",0,1,0 },
	{ "15va down",0,1,0 }
};

static DLOG dg_ksig={ "Insert Key Signature...\n\n",400,8*17,3,Q_OK,Q_ESC,Q_DEL
};
static BTNS bt_ksig1={ CT_RBTN,0,0,0,0,0,2,HZ,8 };
static BTN  rb_ksig1[2]={
	{ "Major",0,1,0 },
	{ "Minor",0,1,0 }
};
static BTNS bt_ksig2={ CT_RBTN,0,0,0,0,0,12,HZ,0 };
static BTN  rb_ksig2[12]={
	{ "C",0,1,0 },
	{ "C#",0,1,0 },
	{ "D",0,1,0 },
	{ "D#",0,1,0 },
	{ "E",0,1,0 },
	{ "F",0,1,0 },
	{ "F#",0,1,0 },
	{ "G",0,1,0 },
	{ "G#",0,1,0 },
	{ "A",0,1,0 },
	{ "A#",0,1,0 },
	{ "B",0,1,0 }
};

static DLOG dg_metr={ "Insert Meter...\n\n",344,8*17,3,Q_OK,Q_ESC,Q_DEL };
static BTNS bt_metr={ CT_RBTN,0,0,0,"Base:",0,6,HZ,8 };
static BTN  rb_metr[6]={
	{ "C",0,1,0 },
	{ "1st",0,1,0 },
	{ "2nd",0,1,0 },
	{ "4th",0,1,0 },
	{ "8th",0,1,0 },
	{ "16th",0,1,0 }
};
static SBAR sb_metr={ CT_SBAR,0,0,0,"Beat Count:",0,HZ,48,16,32,0,2,0,2 };

static DLOG dg_repc={ "Repeat Count...\n\n",256,8*14,2,Q_OK,Q_ESC };
static SBAR sb_repc={ CT_SBAR,0,0,0,"Repeat:",0,HZ,72,16,14,0,2,0,2 };

static DLOG dg_temp={ "Insert Tempo...\n\n",240,8*14,3,Q_OK,Q_ESC,Q_DEL };
static SBAR sb_temp={ CT_SBAR,0,0,0,"Tempo:",0,HZ,80,16,256,0,10,0,3 };

static DLOG dg_octa={ "Octave up/down...\n\n",188,8*22,3,Q_OK,Q_ESC,Q_DEL };
static BTNS bt_octa={ CT_RBTN,0,0,0,0,0,5,VT,4 };

static DLOG dg_gate={ "Gate/Step Ratio...\n\n",200,8*14,3,Q_OK,Q_ESC,Q_DEL };
static SBAR sb_gate={ CT_SBAR,0,0,0," G/S:",0,HZ,32,16,8,0,2,0,1 };

static DLOG dg_lfo={ "LFO on/off...\n\n",160,8*14,3,Q_OK,Q_ESC,Q_DEL };
static BTNS bt_lfo={ CT_RBTN,0,0,0,0,0,2,HZ,8 };
static BTN  rb_lfo[2]={	{ "ON",0,1,0 },{ "OFF",0,1,0 } };

static int  cnvbar[]={ 0,1,12,2,2,9,4,5,8,6 };

extern int  barsz[];
extern int  barskp[];
extern ITEM it_inst[];
extern ITEM it_blck[];

extern char *incbar();
extern char *locbar();

/*** Set Insert Menu ***/

setimnu()
{
	acticon=0;
	if(icsr.tick) acticon|=M_CLEF+M_KSIG+M_TSIG;
	if(part[icsr.part].clef == CLEF_N) acticon|=M_CLEF+M_KSIG;
	if(icsr.part) acticon|=M_TEMPO;
	if(!voicep) acticon|=M_VOICE;
}

/*** Draw Insert Cursor ***/

drawicsr(mode)
int  mode;
{
	ICSR *ip;

	if(mode == INSMODE) ip=&icsr;
	else if(mode == BLKMODE) {
		if(tcsr.part != icsr.part) goto end_icsr;
		ip=&bcsr;
	}
	if(!mode || insblk != mode || ip->part != tcsr.part
	 || ip->bar != tcsr.bar || ip->tick != tcsr.tick) {
		mousoff();
		if(insblk == INSMODE) drawins();
		else if(insblk == BLKMODE && mode != BLKMODE) drawblk(FALSE);
		if(mode == INSMODE) {
			movemem(&tcsr,&icsr,sizeof(ICSR));
			drawins();
			insblk=INSMODE;
		} else if(mode == BLKMODE) {
			if(insblk == INSMODE) {
				movemem(&tcsr,&bcsr,sizeof(ICSR));
				insblk=drawblk(FALSE);
			} else {
				insblk=drawblk(TRUE);
				movemem(&tcsr,&bcsr,sizeof(ICSR));
			}
			if(insblk == INSMODE) drawins();
		}
		mouson();
	}
end_icsr:;
}

/*** Draw Cursor/Block ***/

static drawins()
{	
	pn.col=7;
	pn.op=GP_XOR;
	if(!setcsrxy(&icsr)) line(icsr.px,icsr.py,icsr.px,icsr.py+icsr.sy);
	pn.op=GP_PUT;
}

static int drawblk(f)
flag f;
{
	int  mode;
	int  tx,ty;
	int  ex,ey;
	int  cx;
	int  ret;
	ICSR *tp,*ep;
	
	mode=INSMODE;
	tp=f ? &tcsr : &bcsr;
	if(tp->bar < icsr.bar
	 || tp->bar == icsr.bar && tp->tick <= icsr.tick) {
	 	if(insblk == INSMODE) goto end_dblk;
	 	else { tp=&icsr; ep=&bcsr; }
	} else mode=BLKMODE;
	if(f && mode == BLKMODE) {
		if(tcsr.bar < bcsr.bar
		 || tcsr.bar == bcsr.bar && tcsr.tick <= bcsr.tick) {
			tp=&tcsr; ep=&bcsr;
		} else { tp=&bcsr; ep=&tcsr; }
	} else { tp=&icsr; ep=&bcsr; }
	ret=setcsrxy(tp);
	if(ret > 0) goto end_dblk;
	else if(ret < 0) { tx=SCOR_X; ty=icsr.py; }
	else { tx=tp->px; ty=tp->py; }
	ret=setcsrxy(ep);
	if(ret < 0) goto end_dblk;
	else if( ret > 0) {
		ex=SCOR_X+SCOR_SX;
		ey=icsr.py+stafssy*(mulstaf-1);
	}
	else { ex=ep->px; ey=ep->py; }
	repeat {
		if(ey > ty) cx=SCOR_X+SCOR_SX;
		else cx=ex;
		invrect(tx,ty,cx-tx,icsr.sy);
		tx=SCOR_X;
		ty+=stafssy;
	} until(ty > ey);
end_dblk:
	return(mode);
}

/*** Set Cursor X,Y ***/

static int setcsrxy(ip)
ICSR *ip;
{
	int  i;
	int  n;
	int  m;
	int  x;
	int  xi,yi;
	int  ret;
	PART *pp;

	pp=&part[ip->part];
	n=ip->bar-ndsp;
	m=mb.barlin*mulstaf;
	if(n < 0) ret=-1;
	else if(n >= m) ret=1;
	else {
		xi=n%mb.barlin;
		yi=n/mb.barlin;
		x=SCOR_X;
		for_to(i,xi) x+=barsz[i];
		x+=barskp[(pp->disp-1)*m+n];
		ip->px=x+ip->sumx;
		ip->py=pp->stafy+stafssy*yi;
		ret=0;
	}
	return(ret);
}


/*** Set buttom mode ***/

static onbtn(p,n,c)
BTN  *p;
int  n;
int  c;
{
	int  i;

	for_to(i,c) {
		if(p->mode & MD_INACT) p->mode&=~MD_ON;
		else p->mode=(i == n) ? MD_ON : 0;
		p++;
	}
}

/*** Meter Value ***/

int  valmetr(d)
int  d;
{
	return(d+1);
}

/*** Set Insert Point ***/

int  setinsp(ip,n,data,inspp,seek)
ICSR *ip;
int  n;
int  data;
char **inspp;
flag seek;
{
	int  d0;
	int  tick;
	int  tsum;
	char *insp;
	char *prep;
	char *p,*q;
	flag f;

	if(seek) p=locbar(n,ip->bar);
	else {
		loadprec(n);
		p=incbar(part[n].dspp,ip->bar-ndsp,ndsp);
	}		
	if(data < D_BAR) {
		while(*p == D_BAR) p+=2;
	}
	q=(char*)scanipar(p);
	if(data > D_BAR) {
		while(*q == D_REPT) q+=2;
		p=q;
	}
	prep=p;
	insp=NIL;
	cp.repm=0;
	if(p) {
		tsum=0;
		loop {
			d0=*p;
			if(!d0) break;
			else if(d0 > 0) {
				if(d0 == data) insp=p;
				setiparm(d0,*(p+1));
				if(cp.repm) {
					tsum+=cp.trep*cp.repc;
					cp.repm=0;
					prep=p+2;
				}
			} else {
				tick=settick(d0);
				tsum+=tick;
				cp.trep+=tick;
				if(tsum > ip->tick) break;
				insp=NIL;
				if(cp.chord) p+=2;
				prep=p+2;
			}
			p+=2;
		}
	}
	if(!p) f=-1;
	else {
		if(insp) { *inspp=insp; f=1; }
		else { *inspp=prep; f=0; }
	}
	return(f);
}

/*** Draw Insert Bar ***

static int drawbars(cp)
USER *cp;
{
	int  i;
	int  x,y;
	PEN  pn;
	PEN  *oldpen;

	oldpen=setpen(&pn);
	mousoff();
	x=cp->px;
	y=cp->py;
	pn.col=7;
	micn_s.pat=(char*)meticnp+I_BAR*32;
	for_to(i,INSBARC) {
		fillrect(x,y,PALSZ,PALSZ,NIL);
		shaderct(x,y,PALSZ,PALSZ,1);
		putpat16(x+2,y+2,&micn_s);
		if(i == cp->par) invbars(cp,i);
		x+=PALSZ;
		micn_s.pat+=32;
	}
	mouson();
	respen(oldpen);
	return(y+PALSZ);
}

static invbars(cp,i)
USER *cp;
int  i;
{
	mousoff();
	invrect(cp->px+i*PALSZ,cp->py,PALSZ,PALSZ);
	mouson();
}

static int dobars(cp)
USER *cp;
{
	int  sel;
	int  oldsel;
	int  x,y;
	flag f;

	f=FALSE;
	x=msrec.px-cp->px;
	y=msrec.py-cp->py;
	if(!inrect(x,y,INSBARC*PALSZ,PALSZ)) goto end_do;
	f=TRUE;
	if(msrec.evnt&MS_LDN) {
		sel=x/PALSZ;
		oldsel=cp->par;
		invbars(cp,sel);
		if(sel == oldsel) sel=-1;
		else {
			if(oldsel >= 0) invbars(cp,oldsel);
		}
		cp->par=sel;
	}
end_do:
	return(f);
}

*/

static int cv_repc(val)
int  val;
{
	return(val+3);
}

static int cv_gate(val)
int  val;
{
	return(val+1);
}

/*** Do menu.insert ***/

doinst(icn)
int  icn;
{
	int  i;
	int  dat;
	int  d1,d2;
	int  itm;
	int  insm;
	int  selc;
	int  ret;
	flag allpart;
	char *p;

	if(icn <= IC_BEAMS) {
		itm=D_BEAM;
		d1=icn-IC_BEAMC;
	} else if(icn < IC_CLEF) {
		itm=icn-IC_TEMPO+D_TEMPO;
	} else if(icn < IC_BAR) {
		itm=icn-IC_CLEF+D_CLEF;
	} else {
		itm=D_BAR;
		d1=icn-IC_BAR;
	}	
	if(icsr.bar < ndsp) goto end_inst;
	insm=setinsp(&icsr,icsr.part,itm,&p,FALSE);
	if(insm < 0) goto end_inst;
	if(insm > 0) dat=*(p+1);
	allpart=FALSE;
	ret=Q_OK;
	selc=insm ? 3 : 2;
	switch(itm) {
	case D_CLEF:
		dg_clef.link=(CTRL*)&bt_clef1;
		bt_clef1.link=(CTRL*)&bt_clef2;
		bt_clef1.btnp=&rb_clef1[0];
		bt_clef2.btnp=&rb_clef2[0];
		dg_clef.sel=selc;
		d1=insm ? dat&3 : cp.clef;
		d2=insm ? dat>>2 : cp.clefva;
		if(d1 < 2) {
			onbtn(&rb_clef1[0],d1,2);
		} else {
			rb_clef1[0].mode=rb_clef1[1].mode=MD_INACT;
		}
		onbtn(&rb_clef2[0],2-d2,5);
		ret=dialog(&dg_clef);
		dat=getbtn(&rb_clef1[0],2);
		if(dat < 0) dat=d1;
		dat|=(2-getbtn(&rb_clef2[0],5))<<2;
		break;
	case D_KSIG:
		dg_ksig.link=(CTRL*)&bt_ksig1;
		bt_ksig1.link=(CTRL*)&bt_ksig2;
		bt_ksig1.btnp=&rb_ksig1[0];
		bt_ksig2.btnp=&rb_ksig2[0];
		dg_ksig.sel=selc;
		d1=insm ? dat : cp.ksig;
		d2=0;
		if(d1 & 0x80) { d1&=0x7F; d2=1; }
		onbtn(&rb_ksig1[0],d2,2);
		onbtn(&rb_ksig2[0],d1,12);
		ret=dialog(&dg_ksig);
		dat=(getbtn(&rb_ksig1[0],2)<<7)|getbtn(&rb_ksig2[0],12);
		allpart=TRUE;
		break;
	case D_TSIG:
		dg_metr.link=(CTRL*)&bt_metr;
		bt_metr.link=(CTRL*)&sb_metr;
		bt_metr.btnp=&rb_metr[0];
		dg_metr.sel=selc;
		d1=insm ? dat : cp.meter;
		if(d1) { d2=(d1>>3)&0x1F; d1=d1&7; }
		else { d1=0; d2=3; }
		onbtn(&rb_metr[0],d1,6);
		sb_metr.val=d2;
		sb_metr.func=valmetr;
		ret=dialog(&dg_metr);
		dat=getbtn(&rb_metr[0],6);
		if(dat) dat+=sb_metr.val<<3;
		allpart=TRUE;
		break;
	case D_BAR:
		d2=0;
		if(d1 == 4) {
			dg_repc.link=(CTRL*)&sb_repc;
			sb_repc.val=0;
			sb_repc.func=cv_repc;
			ret=dialog(&dg_repc);
			d2=(sb_repc.val+1)<<4;
		}
		if(ret == Q_OK) {
			d1=cnvbar[d1]|d2;
			if(insm && d1 == dat) ret=Q_DEL;
			dat=d1;
		}
		allpart=TRUE;
		break;
	case D_TEMPO:
		dg_temp.link=(CTRL*)&sb_temp;
		dg_temp.sel=selc;
		sb_temp.val=insm ? (byte)dat : (byte)cp.tempo;
		ret=dialog(&dg_temp);
		dat=sb_temp.val;
		break;
	case D_OCTA:
		dg_octa.link=(CTRL*)&bt_octa;
		bt_octa.btnp=&rb_clef2[0];
		dg_octa.sel=selc;
		d1=insm ? dat : cp.octa;
		onbtn(&rb_clef2[0],2-d1,5);
		ret=dialog(&dg_octa);
		dat=2-getbtn(&rb_clef2[0],5);
		break;
	case D_VOICE:
		ret=selvoic2(&dat,insm);
		break;
	case D_GS:
		dg_gate.link=(CTRL*)&sb_gate;
		dg_gate.sel=selc;
		sb_gate.val=insm ? dat : cp.gate;
		sb_gate.func=cv_gate;
		ret=dialog(&dg_gate);
		dat=sb_gate.val;
		break;
	case D_LFO:
		dg_lfo.link=(CTRL*)&bt_lfo;
		dg_lfo.sel=selc;
		bt_lfo.btnp=&rb_lfo[0];
		d1=insm ? dat : cp.lfo;
		onbtn(&rb_lfo[0],1-d1,2);
		ret=dialog(&dg_lfo);
		dat=1-getbtn(&rb_lfo[0],2);
		break;
	case D_BEAM:
		if(insm && d1 == dat) ret=Q_DEL;
		dat=d1;
		break;
	}
	if(ret != Q_ESC) {
		for_to(i,mb.part) {
			if(allpart || i == icsr.part) {
				insm=setinsp(&icsr,i,itm,&p,FALSE);
				if(insm < 0) continue;
				if(ret == Q_OK) {
					if(itm == D_BAR) {
						if(insm) {
							d2=dat&0x0C;
							loop {
								d1=*(p+1)&0x0C;
								if(d1 == d2) break;
								else if(d1 > d2) { insm=0; break; }
								p+=2;
								if(*p != D_BAR) { insm=0; break; }
							}
						}
					}
					if(insm) *(p+1)=dat;
					else insdata(p,itm,dat);
				}
				else if(insm > 0) deldata(p);
			}
		}
		if(allpart) drawscor();
		else drawstf1(icsr.part);
	}
end_inst:
}
