/***************************
	'MET' Draw Note etc.
***************************/

#include "std.h"
#include "graf.h"
#include "met.h"

#define AMAX	192

#define P_NATU	0x08*64
#define P_SHARP 0x09*64
#define P_WSHAP	0x0A*64
#define P_FLAT	0x0B*64
#define P_WFLAT	0x0C*64
#define P_REST	0x18*64
#define P_RESTA	P_REST+48
#define P_STEM	0x10*64
#define P_NOTE	0x1D*64
#define P_DOT	P_NOTE+48
#define P_NOIS	0x1E*64
#define P_STAC	0x1F*64
#define P_ACC	P_STAC+48
#define P_PORT	0x26*64+32

#define RESTWD	10

typedef struct restrec {
	int  patp;
	int  py;
	int  sy;
} REST;

typedef struct beamrec {
	int  x;
	int  y;
	int  gyh;
	int  gyl;
	int  len;
	int  by;
	int  acc;
	flag key;
} BEAM;

extern BEAM beamwk[64];
extern BEAM	*bmp;
extern int  bmi;
extern int ticks[];

static int  kpat[12]={
	 P_NATU,P_SHARP,P_WSHAP,0,
	 P_WFLAT,P_FLAT,P_NATU,0,
	 P_FLAT,P_NATU,P_SHARP,0
};
static REST rests[8]={
	{ P_RESTA,9,4 },
	{ P_RESTA,9,4 },
	{ P_RESTA,12,4 },
	{ P_REST,4,23 },
	{ P_REST+64,9,15 },
	{ P_REST+64*2,9,23 },
	{ P_REST+64*3,1,31 },
	{ P_REST+64*4,1,31 }
};

/*** Draw note ***/

int  drawnote(px,py,kdat,ldat,stem,acc)
int  px,py;
int  kdat;
int  ldat;
flag stem;
flag acc;
{
	int  i;
	int  va;
	int  leger;
	int  oct;
	int  tone;
	int  len;
	int  key;
	int  gy;
	int  y;
	int  p;
	int  gpy;
	int  dot;
	REST *rp;

	gpy=py+cp.vtmgn;
	dot=ldat & 0x18;
	if(dot && dot != 0x18) dot=P_DOT+((dot == 0x10) ? 8 : 0);
	else dot=0;	
	if(kdat == RESTC) {
		if(settick(ldat) == cp.tmesr) {
			px+=(mesrsx-RESTWD)/2;
		}
		rp=&(rests[ldat & 0x07]);
		putmpat(px,gpy+rp->py,rp->sy,rp->patp);
		if(dot) {
			putmpat(px+11,gpy+12,2,dot);
		}
		gy=0;
	} else {
		va=0;
		leger=0;
		tone=kdat & 0x07;
		oct=(kdat & 0x38)>>3;
		key=oct*7+tone;
		while(key > cp.key_h+cp.legmax) {
			key-=7;
			va++;
		}
		while(key < cp.key_l-cp.legmax) {
			key+=7;
			va--;
		}
		if(key > cp.key_h) leger=(key-cp.key_h+1)/2;
		else if(key < cp.key_l) leger=-((cp.key_l-key+1)/2);
		gy=cp.key_h-key-1;
/* Draw Leger */
		if(leger > 0) {
			y=gpy;
			for_to(i,leger) {
				y-=GRID;
				legline(px,y,ldat);
			}
		} else if(leger < 0) {
			y=gpy+cp.gridsy;
			for_to(i,-leger) {
				y+=GRID;
				legline(px,y,ldat);
			}
		} else if(gy == 10) {
			y=gpy+GRID*5;
			legline(px,y,ldat);
		}
/* Draw va */
		pn.lin=0xCCCC;
		if(va > 0) {
			y=py;
			for_to(i,va) {
				y+=VASPACE;
				legline(px,y,ldat);
			}
		}
		else if(va < 0) {
			y=py+cp.stafsy;
			for_to(i,-va) {
				y-=VASPACE;
				legline(px,y,ldat);
			}
		}
		pn.lin=0xFFFF;
/* Put Note */
		y=gpy+gy*(GRID/2);
		if(ldat) {
			len=ldat & 0x07;
			if(cp.clef == CLEF_N) {
				p=(len <= LEN_2) ? 0 : 32;
				putmpat(px,y-4,9,P_NOIS+p);
			} else {
				switch(len) {
				case LEN_A:
				case LEN_1:
					p=0;
					break;
				case LEN_2:
					p=16;
					break;
				default:
					p=32;
					break;
				}
				putmpat(px,y-3,8,P_NOTE+p);
			}
		} else {
			putmpat(px,y-2,6,P_PORT);
		}
/* Put Dot */
		if(dot) {
		
			putmpat(px+11,y-((gy&1) ? 1 : 3),2,dot);
		}
/* Put #/b */
		if(kdat & 0xC0) {
			p=kpat[((kdat>>4)&0x0C)-3+krefb[tone]];
			drawshap(px-9,y,p,ldat);
		}
/* Put Stem */
	}
	bmp->x=px;
	bmp->gyh=bmp->gyl=gy;
	bmp->len=ldat;
	bmp->key=kdat != RESTC;
	bmp->acc=acc;
	if(stem) drawbeam(py);
	return(gy);
}

static legline(x,y,type)
int  x,y;
flag type;
{
	if(type) line(x-3,y,x+10,y);
	else line(x-2,y,x+8,y);
}

/*** Draw Chord ***/

int  drawcord(px,py,cdatp,ldat,stem,acc)
int  px,py;
char *cdatp;
int  ldat;
flag stem;
flag acc;
{
	int  i;
	int  kdat;
	int  gy;
	int  gyh,gyl;
	flag key;
	char *p;

	key=TRUE;
	p=cdatp;
	for_to(i,3) {
		kdat=*p++;
		if(kdat == RESTC && i) break;
		gy=drawnote(px,py,kdat,ldat,FALSE);
		if(kdat == RESTC) {
			stem=key=FALSE;
			break;
		}
		if(!i) gyh=gyl=gy;
		else {
			if(gy < gyh) gyh=gy;
			if(gy > gyl) gyl=gy;
		}
	}
	bmp->gyh=gyh;
	bmp->gyl=gyl;
	bmp->len=ldat;
	bmp->key=key;
	bmp->acc=acc;
	if(stem) drawbeam(py);
	return(gy);
}

/*** Draw Beam ***/

flag drawbeam(py)
int  py;
{
	int  i,j,k;
	int  y,by;
	int  x1,y1;
	int  x2,y2;
	int  sy;
	int  dx,dy;
	int  a,a1;
	int  hi,li,ei;
	int  gy,gyh,gyl;
	int  kyh,kyl;
	int  tsum;
	int  adx,ady;
	int  tedge;
	int  bmc;
	BEAM *bp,*bp1,*bp2;
	flag beam;
	flag stemup;
	flag f;
	flag bmf;

	if(!bmi) bmi++;
	bp=&beamwk[0];
	gyh=0x7FFF;
	gyl=0x8000;
	if(!bp->key) goto end_beam;
	bmc=0;
	for_to(i,bmi) {
		if(bp->key) {
			if(bp->gyh < gyh) { gyh=bp->gyh; hi=i; }
			if(bp->gyl > gyl) { gyl=bp->gyl; li=i; }
			bmc=i;
		}
		bp++;
	}
	bmi=bmc+1;
	if(cp.clef == CLEF_GF) {
		kyh=(gyh > 10) ? 11 : 0;
		kyl=(gyl > 10) ? 20 : 8;
	} else {
		kyh=0;
		kyl=8;
	}
	stemup=(gyh-kyh > kyl-gyl);
	bp=&beamwk[0];
	for_to(i,bmi) {
		bp->y=py+cp.vtmgn+(stemup ? bp->gyl : bp->gyh)*(GRID/2);
		bp++;
	}
	beam=FALSE;
	ei=0;
	if(bmi > 1) {
		a=0;
		ei=stemup ? hi : li;
		tsum=0;
		if(ei == 0 || ei == bmi-1) {
			bp=ei ? &beamwk[bmi-1] : &beamwk[0];
			a=0x7FFF;
			for_to(i,bmi-1) {
				if(ei) bp--;
				tsum+=ticks[bp->len&0x1F];
				if(!ei)	bp++;
				if(bp->key) {
					if(stemup) a1=sftdiv((bp->gyh-gyh),tsum);
					else a1=sftdiv((gyl-bp->gyl),tsum);
					if(a1 < a) a=a1;
				}
			}
			if(a > AMAX) a=AMAX;
			a*=(GRID/2);
		}
		beam=TRUE;
	}
	gy=(stemup) ? beamwk[ei].gyh : beamwk[ei].gyl;
	by=py+cp.vtmgn+gy*(GRID/2);
	if(stemup) {
		by-=STEMLN;
		if(by < py) by=py;
	} else {
		by+=STEMLN;
		if(by > py+cp.stafsy) by=py+cp.stafsy;
	}
	bp=&beamwk[0];
	if(beam) {
		tsum=0;
		bp=ei ? &beamwk[bmi-1] : &beamwk[0];
		for_to(i,bmi) {
			if(bp->key) {
				dy=sftmul(a,tsum);
				sy=bp->by=(stemup) ? by+dy : by-dy;
				y=bp->y;
				sy=(stemup) ? y-sy : sy-y;
				drawstem(bp->x,y,sy,LEN_2,stemup);
			}
			if(ei) bp--;
			tsum+=ticks[bp->len&0x1F];
			if(!ei)	bp++;
		}
		for_to(i,4) {
			bmc=0;
			bp1=NIL;
			bp=&beamwk[0];
			bmf=FALSE;
			for_to(j,bmi) {
				if(bp->key) {
					beam=FALSE;
					if((bp->len&7) >= i+4) {
						if(!bmc) bp1=bp;
						bp2=bp;
						bmc++;
						if(j == bmi-1) beam=TRUE;
						bmf=TRUE;
					} else {
						if(bmc) beam=TRUE;
					}
					if(beam) {
						x1=bp1->x;
						y1=bp1->by;
						x2=bp2->x;
						y2=bp2->by;
						if(bmc == 1) {
							f=stemup;
							if(ei) f=!f;
							adx=sumtox(6);
							ady=sftmul(a,6);
							if(bp1 == &beamwk[0]) {
								if((tedge=bp1->len&7) >= 6) {
									tedge=(8-tedge)*2;
									adx=sumtox(tedge);
									ady=sftmul(a,tedge);
								}
								x2=x1+adx;
								f=!f;
							} else x2=x1-adx;
							y2=f ? y1-ady : y1+ady;
						}
						dy=i*6;
						if(stemup) {
							dx=7;
						} else {
							dx=0;
							dy=-dy-3;
						}
						if(cp.clef == CLEF_N) dx=4;
						for_to(k,4) {
							line(x1+dx,y1+dy,x2+dx,y2+dy);
							dy+=1;
						}
						bmc=0;
						bp1=NIL;
					}
				}
				bp++;
			}
			if(!bmf) break;
		}
	} else {
		bp->by=by;
		y=bp->y;
		sy=(stemup) ? y-by : by-y;
		drawstem(bp->x,y,sy,bp->len,stemup);
	}
	bp=&beamwk[0];
	for_to(i,bmi) {
		if(bp->key) {
			if(bp->len&0x60) {
				drawstac(bp);
			}
			if(bp->acc) drawacc(bp,py);
		}
		bp++;
	}
end_beam:
	bmi=0;
	bmp=&beamwk[0];
}

/*** Draw Stem ***/

static drawstem(px,py,sy,len,stemup)
int  px,py;
int  sy;
int  len;
flag stemup;
{
	int  x,y;
	int  p;

	len&=7;
	if(stemup) {
		x=px+7;
		y=py-sy;
	} else {
		x=px;
		y=py+sy;
	}
	if(cp.clef == CLEF_N) x=px+4;
	if(len > LEN_1) line(x,py,x,y);
	if(len >= LEN_8) {
		p=P_STEM+(len-LEN_8)*64;
		if(!stemup) {
			y-=32;
			p+=64*4;
		}
		putmpat(x,y,32,p);
	}
}

/*** Draw Staccato ***/

static drawstac(bp)
BEAM *bp;
{
	int  dy;
	int  sy;
	int  p;
	int  gs;
	int  ldat;
	int  stmy;

	ldat=bp->len;
	stmy=bp->y-bp->by;
	if((ldat&7) < 2) stmy=0;
	gs=(ldat&0x60)>>5;
	if(gs) {
		sy=4;
		if(gs == 1) {
			if(stmy >= 0) { p=32; dy=6; }
			else { p=40; dy=-10; }
		} else {
			if(gs == 2) p=0;
			else { sy=6; p=16; }
			dy=((stmy > 0) ? -stmy : -4)-sy;
		}
		putmpat(bp->x+2,bp->y+dy,sy,P_STAC+p);
	}
}

/*** Draw Accento ***/

static drawacc(bp,py)
BEAM *bp;
int  py;
{
	int  y;

	y=py+cp.vtmgn;
	if(bp->y < y) y=bp->y;
	if(bp->by < y) y=bp->by;
	putmpat(bp->x,y-12,8,P_ACC);
	bp->acc=FALSE;
}
