/*************************
	'MET' Part Handler
*************************/

#include "std.h"
#include "graf.h"
#include "met.h"

#define I_TEMPO	120
#define M_CLEF	0x03
#define M_CHORD	0x0C
#define M_FM	0x80
#define CD_NORM	0
#define CD_2CH	1
#define CD_3CH	2
#define CD_FM3	3
#define P_CLEF	0
#define P_CHORD 3
#define ESCAPE	-2
#define IC_NOTE	0

static char nm_fm[]="FM";
static char nm_psg[]="PSG";
static char nm_ch3[]="ch3";
static char nm_nois[]="Noise";

static DLOG dg_part={ "Select Part...\n\n",320,8*14,2,Q_OK,Q_ESC };
       BTNS bt_part={ CT_PBTN,0,0,0,0,0,0,HZ,8 };
       BTN  pb_part[7];
static DLOG dg_ptyp={ "Part Type...\n\n",320,8*14,2,Q_OK,Q_ESC };
static BTNS bt_ptyp={ CT_RBTN,0,0,0,0,0,3,HZ,8,0 };
static BTN  rb_ptyp[3]={
	{ "FM ",0,1,0 },
	{ "PSG",0,1,0 },
	{ "Noise",0,1,0 }
};
static DLOG dg_pform={ NIL,330,8*14,2,Q_OK,Q_ESC };
static BTNS bt_pform[2]={ 
	{ CT_RBTN,0,0,0,"Clef:",64,3,HZ,8,0 },
	{ CT_RBTN,0,0,0,"Chord:",64,4,HZ,8,0 }
};
static BTN  rb_pform[7]={
	{ "G",0,2,0 },
	{ "F",0,2,0 },
	{ "G&F",0,2,0 },
	{ "Single",0,3,0 },
	{ "Double",0,3,0 },
	{ "Triple",0,3,0 },
	{ "FM3",0,3,0 }
};
static DLOG dg_scor={ "Setup Score...\n\n",240,8*23,2,Q_OK,Q_ESC };
static BTNS bt_bars={ CT_RBTN,0,0,0,"Bars/Line:",80,4,HZ,8,0 };
static BTN  rb_bars[4]={
	{ "1",0,1,0 },
	{ "2",0,1,0 },
	{ "3",0,1,0 },
	{ "4",0,1,0 },
};
static BTNS bt_numb={ CT_RBTN,0,0,0,"Measure Numbers: ",0,1,HZ,8 };
static BTN  rb_numb={ "",0,0,0 };
static BTNS bt_mini={ CT_RBTN,0,0,0,"Minimum Time: ",0,2,HZ,8 };
static BTN  rb_mini[2]={ { "32th",0,1,0 },{ "64th",0,1,0 } };
static SBAR sb_beat={ CT_SBAR,0,0,0,"Beat Shift:",80,HZ,48,16,0,0,2,0,2 };

static DLOG dg_staf={ "Setup Part...\n\n",240,8*18,2,Q_OK,Q_ESC };
static SBAR sb_marg={ CT_SBAR,0,0,0,"V space:",80,HZ,48,16,0,0,2,0,2 };

extern DLOG dg_full;
extern int  tickc[];

/*** Part Block Handler ***/

PBLK *pblkptr(n)
int  n;
{
	int  i;
	PBLK *p;

	p=(PBLK*)(musicp+sizeof(MBLK));
	for_to(i,n) p=(PBLK*)((char*)p+p->size);
	return(p);
}
	
setpart(n,mode)
int  n;
int  mode;
{
	PBLK *p;

	p=pblkptr(n);
	p->mode=part[n].mode=mode;
	initpart();
}

/*** Make Part ***/

int  makepart(n,mode)
int  n;
int  mode;
{
	int  f;
	PBLK *p;

	if(n >= 0) p=pblkptr(n); else p=(PBLK*)freep;
	if(f=movepart(p,p+1)) {
		p->size=sizeof(PBLK);
		p->mode=mode;
		p->vmgn=4;
		p->data=0;
		if(n >= 0 ) {
			if(mb.part) {
				movemem(&part[n],&part[n+1],(mb.part-n)*sizeof(PART));
			}
		} else n=mb.part;
		initprec(n);
		part[n].barc=0;
		mb.part++;
		initpart();
	}
	return(f);
}

/*** Remove Part ***/

int rmovpart(n)
int  n;
{
	PBLK *p;

	p=pblkptr(n);
	movepart((char*)p+p->size,p);
	movemem(&part[n+1],&part[n],(mb.part-n-1)*sizeof(PART));
	mb.part--;
	initpart();
	return(OK);
}

/*** Insert/Delete Block ***/

int  inspart(pp,p,dp,size)
PBLK *pp;
char *p;
char *dp;
int  size;
{
	int  f;

	if(f=movepart(p,p+size)) {
		movemem(dp,p,size);
		pp->size+=size;
	}
	return(f);
}

int  delpart(pp,p,size)
PBLK *pp;
char *p;
int  size;
{
	char *endp;

	endp=(char*)pp+pp->size-2;
	if(p+size > endp) {
		if(p < endp) size=endp-p;
		else size=0;
	}
	pp->size-=size;
	return(movepart(p+size,p));
}

/*** Insert/Delete Data ***/

int  insdata(p,d0,d1)
char *p;
char d0;
char d1;
{
	int  f;

	if(f=movepart(p,p+2)) {
		*p++=d0;
		*p++=d1;
		(cp.partp)->size+=2;
	}
	return(f);
}

int  deldata(p)
char *p;
{
	delpart(cp.partp,p,2);
}

/*** Move Part Block ***/

static int movepart(sp,dp)
char *sp,*dp;
{
	int  ofst;
	int  f;

	ofst=dp-sp;
	if(freep+ofst > clipp) {
		dialog(&dg_full);
		f=ERROR;
	} else {
		movemem(sp,dp,freep-sp);
		ptradj(sp,ofst);
		f=OK;
	}
	return(f);
}
	
/*** Pointer adjust ***/

ptradj(mp,ofst)
char *mp;
int  ofst;
{
	int  i,j;
	word *p;

	freep+=ofst;
	for_to(i,mb.part) {
		p=(word*)&part[i];
		for_to(j,PARTPC) {
			if(*p) {
				if(*p > (word)mp) *p+=ofst;
				else if(*p > (word)(mp+ofst)) *p=(word)(mp+ofst);
			}
			p++;
		}
	}
}

/*** Init Music record ***/

initmusc()
{
	ndsp=0;
	calcbar();
	initpart();
}

/*** Init Part record ***/

initprec(n)
int  n;
{
	int  mode;
	PART *pp;
	PBLK *bp;
	char *cp;

	pp=&part[n];
	fillmem((char*)pp+PARTRH,sizeof(PART)-PARTRH,0);
	bp=pp->partp=pblkptr(n);
	pp->mode=mode=bp->mode;
	pp->legmax=bp->vmgn;
	cp=(char*)&(bp->data);
	pp->topp=pp->dspp=pp->nowp=cp;
	pp->tempo=I_TEMPO;
	pp->chord=(mode&M_CHORD)>>2;
	pp->dyna=4;
	pp->gate=6;
	pp->lfo=1;
	initsize(pp);
}

/*** Init size ***/

static initsize(pp)
PART *pp;
{
	pp->vtmgn=setvmgn(pp->legmax);
	pp->clef=pp->mode&M_CLEF;
	pp->gridsy=(pp->clef == CLEF_GF) ? GRID*10 : GRID*4;
	pp->stafsy=pp->gridsy+pp->vtmgn*2;
}

/*** Init Part ***/

static initpart()
{
	int  i,j;
	int  mode;
	int  use;
	char c;
	char *p;
	flag fm;

	usefm=usefm3=usepsg=usenois=0;
	for_to(i,mb.part) {
		mode=part[i].mode;
		if((mode&M_CLEF) == CLEF_N) {
			strcpy(part[i].name,nm_nois);
			usenois++;
		} else {
			fm=!(mode&M_FM);
			p=fm ? nm_fm : nm_psg;
			p=strcpy(part[i].name,p);
			use=((mode&M_CHORD)>>2)+1;
			if(use < 4) {
				if(fm) { c=usefm; usefm+=use; }
				else { c=usepsg; usepsg+=use; }
				for_to(j,use) *p++='1'+c++;
			} else {
				usefm++;
				usefm3++;
				p=strcpy(p,nm_ch3);
			}
			*p=0;
		}
	}
	mute=0;
	imode=insblk=0;
	icon=IC_NOTE;
	initpsel();
}

/*** Add/Del Staff ***/

int  addstaf()
{
	int  n;
	int  mode;
	int  f;

	f=ERROR;
	if((n=selpart()) != ESCAPE) {
		if((mode=setptyp(0)) != ESCAPE) {
			f=makepart(n,mode);
		}
	}
	return(f);
}

int  delstaf()
{
	int  n;
	int  f;

	f=ERROR;
	if((n=selpart()) >= 0) f=rmovpart(n);
	return(f);
}

/*** Select Part ***/

static int selpart()
{
	int  i;
	int  f;

	f=-1;
	if(mb.part) {
		setbtns(0);
		dg_part.link=(CTRL*)&bt_part;
		bt_part.px=bt_part.py=0;
		if(dialog(&dg_part) == Q_ESC) f=ESCAPE;
		else {
			for_to(i,mb.part) {
				if(pb_part[i].mode & MD_ON) {
					f=i;
					break;
				}
			}
		}
		setbtns(1);
	}
	return(f);
}	

/*** Set Buttons ***/

setbtns(mode)
int mode;
{
	int  i;
	int  btnmod;
	int  btst;

	bt_part.btnp=&pb_part[0];
	bt_part.btns=mb.part;
	btst=1;
	for_to(i,mb.part) {
		pb_part[i].ttl=part[i].name;
		pb_part[i].n=!mode;
		btnmod=MD_BTN;
		switch(mode) {
		case 0:
			break;
		case 1:
			if(part[i].disp) btnmod= MD_ON+MD_BTN;
			break;
		case 2:
			if(!(mute&btst)) btnmod=MD_ON+MD_BTN;
			break;
		}
		pb_part[i].mode=btnmod;
		btst<<=1;
	}
}

/*** Set Part Type ***/

static int setptyp(mode)
int  mode;
{
	int  i;
	int  type;
	int  use;

	dg_ptyp.link=(CTRL*)&bt_ptyp;
	bt_ptyp.btnp=&rb_ptyp[0];
	dg_pform.link=(CTRL*)&bt_pform[0];
	bt_pform[0].link=(CTRL*)&bt_pform[1];
	bt_pform[0].px=bt_pform[0].py=0;
	bt_pform[0].btnp=&rb_pform[P_CLEF];
	bt_pform[1].btnp=&rb_pform[P_CHORD];
	if((mode&M_CLEF) == CLEF_N) i=2;
	else i=(mode&M_FM) ? 1 : 0;
	rb_ptyp[0].mode=(usefm == 3) ? MD_INACT :0;
	rb_ptyp[1].mode=(usepsg == 3) ? MD_INACT :0;
	rb_ptyp[2].mode=usenois ? MD_INACT :0;
	setbtn(&rb_ptyp[0],i,3);
	if(dialog(&dg_ptyp) == Q_ESC) mode=ESCAPE;
	else {
		if((type=getbtn(&rb_ptyp[0],3)) == ESCAPE) mode=ESCAPE;
		else if(type == 2) mode=CLEF_N;
		else {
			setbtn(&rb_pform[P_CLEF],mode & M_CLEF,3);
			use=type ? usepsg : usefm;
			for_to(i,3) {
				rb_pform[P_CHORD+i].mode=(i >= 3-use) ? MD_INACT :0;
			}	
			rb_pform[P_CHORD+3].mode
				=(usefm == 3 || usefm3 || type) ? MD_INACT : 0;
			setbtn(&rb_pform[P_CHORD],((mode & M_CHORD) >> 2),4);
			if(dialog(&dg_pform) == Q_ESC) mode=ESCAPE;
			else {
				mode=type<<7;
				mode|=getbtn(&rb_pform[P_CLEF],3);
				mode|=getbtn(&rb_pform[P_CHORD],4)<<2;
			}
		}
	}
	return(mode);
}

static setbtn(p,n,c)
BTN  *p;
int  n;
int  c;
{
	int  i;

	for_to(i,c) {
		if(i == n) {
			if(p->mode&MD_INACT) n++;
			else p->mode|=MD_ON;
		}
		else {
			if(!(p->mode&MD_INACT)) p->mode=0;
		}
		p++;
	}
}

int getbtn(p,c)
BTN  *p;
int  c;
{
	int  i;
	int  r;

	r=ESCAPE;
	for_to(i,c) {
		if(p->mode & MD_ON) {
			r=i;
			break;
		}
		p++;
	}
	return(r);
}

/*** Setup Score ***/

int  setscor()
{
	int  f;

	f=ERROR;
	dg_scor.link=(CTRL*)&bt_bars;
	bt_bars.btnp=&rb_bars[0];
	bt_bars.link=(CTRL*)&bt_numb;
	bt_numb.btnp=&rb_numb;
	bt_numb.link=(CTRL*)&bt_mini;
	bt_mini.btnp=&rb_mini[0];
	bt_mini.link=(CTRL*)&sb_beat;
	setbtn(&rb_bars[0],mb.barlin-1,4);
	setbtn(&rb_mini[0],!!tickc[1],2);
	rb_numb.mode=mesrnum ? MD_ON : 0;
	sb_beat.val=mb.beat;
	sb_beat.max=cp.meter ? (cp.meter>>3)+1 : 4;
	if(dialog(&dg_scor) == Q_OK) {
		mb.barlin=getbtn(&rb_bars[0],4)+1;
		tickc[1]=getbtn(&rb_mini[0],2) ? 3 :0;
		mb.beat=sb_beat.val;
		mesrnum=!!rb_numb.mode;
		f=OK;
	}
	return(f);
}

/*** Setup Staff ***/

int  setstaf()
{
	int  n;
	int  f;
	PBLK *p;
	PART *pp;

	f=ERROR;
	if((n=selpart()) >= 0) {
		pp=&part[n];
		p=pp->partp;
		dg_staf.link=(CTRL*)&bt_pform[0];
		bt_pform[0].btnp=&rb_pform[P_CLEF];
		bt_pform[0].px=bt_pform[0].py=0;
		bt_pform[0].link=(CTRL*)&sb_marg;
		setbtn(&rb_pform[P_CLEF],p->mode&M_CLEF,3);
		sb_marg.val=p->vmgn;
		sb_marg.max=32;
		if(dialog(&dg_staf) == Q_OK) {
			p->mode&=~M_CLEF;
			p->mode|=getbtn(&rb_pform[P_CLEF],3);
			pp->mode=p->mode;
			pp->legmax=p->vmgn=sb_marg.val;
			initsize(pp);
			initpsel();
			f=OK;
		}
	}
	return(f);
}

static int setvmgn(x)
int  x;
{
	return(x*(GRID/2)+GRID+MSG_SY);
}

/*** Init Part Select ***/

initpsel()
{
	int  i;
	int  sy;
	int  dspc;
	PART *pp;
	PART *dp;

	sy=0;
	dspc=0;
	dp=NIL;
	pp=&part[0];
	dsppart=-1;
	for_to(i,mb.part) {
		pp->stafsy=pp->gridsy+pp->vtmgn*2;
		if(SCOR_SY-sy >= pp->stafsy) {
			pp->disp=++dspc;
			pp->dspbtm=FALSE;
			pp->stafy=SCOR_Y+sy;
			sy+=pp->stafsy;
			dp=pp;
			if(dsppart < 0) dsppart=i;
		} else pp->disp=0;
		pp++;
	}
	if(dp) dp->dspbtm=TRUE;
	stafssy=sy;
	setmstaf();
}

setmstaf()
{
	int  sy;

	mulstaf=1;
	if(sy=stafssy) {
		while(SCOR_SY-sy > stafssy) {
			sy+=stafssy;
			mulstaf++;
		}
	}
	scorsy=stafssy*mulstaf;
	setimnu();
}

/*** Voice number adjust ***/

vnumadj(n,d)
int  n;
int  d;
{
	int  i,j;
	int  d0,d1;
	MBLK *mp;
	PBLK *pp;
	char *p;
	flag chord;

	mp=(MBLK*)(mheadp+1);
	for_to(i,mcount+1) {
		pp=(PBLK*)(mp+1);
		for_to(j,mp->part) {
			chord=(pp->mode&M_CHORD)>>2;
			p=(char*)pp+4;
			while(d0=*p) {
				if(d0 < 0) {
					if(chord) p+=2;
				} else if(d0 == D_VOICE) {
					d1=*(p+1);
					if(d1 > n || d1 == n && d > 0) *(p+1)=d1+d;
				}
				p+=2;
			}
			pp=(PBLK*)((char*)pp+pp->size);
		}
		mp=(MBLK*)((char*)mp+mp->size);
	}
}
