/**********************
	Field Map Editor
**********************/

#include "std.h"
#include "graf.h"
#include "pet.h"

#define MENUS	3
#define MN_LOD	0
#define MN_SAV	1
#define MN_NEW	2
#define MN_END	3
#define MN_VEW	0
#define MN_EDT	1
#define MN_CPY	2
#define MN_ARA	3
#define MN_PNT	4
#define MN_FRM	0
#define	MN_SIZ	1

#define	DR_RGT	0x01
#define	DR_LFT	0x02
#define DR_UP	0x04
#define DR_DWN	0x08

#define VBX_X	1*8
#define VBX_Y	4*8
#define VBX_SX	16
#define VBX_SY  10
#define SCR_X0	68*8
#define SCR_Y0	4*8
#define SCR_X1	39*8
#define SCR_Y1  14*8
#define ANUMSX	80

#define FLDMAX	2048
#define MAX_SX	64
#define MAX_SY	64
#define ATBLSZ	128
#define FPMAX	26	

typedef struct fmaprec {
	byte spr;
	byte n;
	word obst;
} FMAP;

typedef struct treerec {
	byte n1;
	byte n2;
} TREE;

typedef struct fptrrec {
	int  fpx;
	int  fpy;
} FPTR;

typedef struct header {
	char id[4];
	word size;
	int  fsx;
	int  fsy;
	int  areac;
	int  fptrc;
	int	 nul;
} HEAD;

extern int  bgobst;
extern flag fllmap;
extern MBAR mb_selw;
extern DLOG dg_sure;
extern DLOG dg_lodng;
extern TREE *chknp();
extern int  cv_base();

FMAP *fmaptop;
TREE *areatbl;
FPTR *fptrtop;
int  atbl_c;
int  fptr_c;
word fmapbyte;
word fsx=16;
word fsy=16;
word fpx;
word fpy;
flag fullmap;

static word view_x;
static word view_y;
static word view_sx;
static word view_sy;
static word scrl_x;
static word scrl_y;
static int  fmapmode;
static flag frame;
static char fmapname[]="untitled.fmp";
static char fmapmask[]="*.fmp";
static char fmapid[]="FMP";
static int  cp_x1;
static int  cp_y1;
static int  cp_x2;
static int  cp_y2;
static int  selarea;
static FMAP *fmapp;

static MENU mn_fmap[]={
	"File",4,80,0,0,0,
	"mode",5,80,0,0,0,
	"Format",2,80,0,0,0
};
static ITEM it_file[]={
	"Load",0,
	"Save",0,
	"New",0,
	"Quit",0
};
static ITEM it_mode[]={
	"View",0,
	"Edit",0,
	"Copy",0,
	"Area",0,
	"Point",0
};
static ITEM it_form[]={
	"Frame",0,
	"Size",0
};

static int  pt_map[4]={ 0xEEEE,0xEEEE,0xEEEE,0x0000 };
static PATT map_s={0,2,4,1};

static DLOG dg_size={ "Field Size:\n\n",200,8*17,2,Q_OK,Q_ESC };
static DLOG dg_area={ "New Area No.:\n\n",248,8*14,2,Q_OK,Q_ESC };
static SBAR sb_count={ CT_SBAR,0,0,0,NIL,0,HZ,128+16,16,128,0,10,0,3 };
static SBAR sb_sizex={ CT_SBAR,0,0,0,"X:",0,HZ,64+16,16,MAX_SX,0,8,0,2 };
static SBAR sb_sizey={ CT_SBAR,0,0,0,"Y:",0,HZ,64+16,16,MAX_SY,0,8,0,2 };

/*** Draw view box ***/

static drawview()
{
	word x,y;
	int  f;
	FMAP *mp;
	
	mousoff();
	f=frame ? 2 : 0;
	mp=fmaptop+fpy*fsx+fpx;
	for_to(y,view_sy) {
		for_to(x,view_sx) {
			if(fmapmode == MN_ARA) {
				f=inarea(mp->n,selarea) ? 2 : 0;
			}
			putfspr(x,y,mp->spr,f);
			mp++;
		}
		mp+=fsx-view_sx;
	}
	if(fmapmode == MN_ARA) drawarea(selarea);
	mouson();
}

/*** Put sprite ***/

putfspr(x,y,spr,f)
word x,y;
int  spr;
int  f;
{
	mousoff();
	x=x*32+view_x;	
	y=y*32+view_y;
	putzspr(x,y,patptr,3,0,spr*16+mapptr);
	if(f) {
		pn.col=f;
		framerct(x,y,32,32);
	}
	mouson();
}

/*** Do File ***/

static dofile(itm)
int  itm;
{
	switch(itm) {
	case MN_LOD:
		loadfmap();
		initfmap();
		drawall();
		break;
	case MN_SAV:
		savefmap();
		break;
	case MN_NEW:
		if(dialog(&dg_sure)) {
			freemem(fmapptr);
			fmapptr=NIL;
			initfmap();
			drawfmap();
		}
		break;
	case MN_END:
		fldmap=FALSE;
		init();
		break;
	}
}

/*** Do Mode ***/

static domode(itm)
int itm;
{
	int  oldmode;
	
	if(itm != fmapmode) {
		oldmode=fmapmode;
		fmapmode=itm;
		setimark(&it_mode[0],5,fmapmode);
		if(itm == MN_VEW || oldmode == MN_VEW) {
			setfmap();
			drawall();
		}
		if(fmapmode == MN_ARA) drawview();
	}
}
		

/*** Do Format ***/

static doform(itm)
int  itm;
{
	int  sx;
	int  sy;

	switch(itm) {
	case MN_FRM:
		frame=!frame;
		drawview();
		break;
	case MN_SIZ:
		dg_size.link=(CTRL*)&sb_sizex;
		sb_sizex.link=(CTRL*)&sb_sizey;
		sb_sizex.val=fsx-1;
		sb_sizey.val=fsy-1;
		sb_sizex.func=sb_sizey.func=cv_base;
		if(dialog12(&dg_size) == Q_OK) {
			sx=sb_sizex.val+1;
			sy=sb_sizey.val+1;
			if(sx*sy <= FLDMAX) {
				fsx=sx;
				fsy=sy;
				freemem(fmapptr);
				fmapptr=NIL;
				initfmap();
				drawfmap();
			}
		}
	}
}

/*** Init Fmap ***/

initfmap()
{
	int mpsz;
	
	mb_fmap.mnu=&mn_fmap[0];
	mb_fmap.ttl=fmapname;	

	mn_fmap[0].item=it_file;
	mn_fmap[1].item=it_mode;
	mn_fmap[2].item=it_form;
	mn_fmap[0].func=&dofile;
	mn_fmap[1].func=&domode;
	mn_fmap[2].func=&doform;
	setimark(&it_mode[0],5,fmapmode);

	if(!fmapptr) {
		mpsz=fsx*fsy;
		fmapbyte=mpsz*sizeof(FMAP)+HEADSZ+ATBLSZ*sizeof(TREE)+FPMAX*sizeof(FPTR);
		fmapptr=getmem(fmapbyte);
		fillmem(fmapptr,fmapbyte,0);
		fmaptop=(FMAP*)(fmapptr+HEADSZ);
		areatbl=(TREE*)(fmaptop+mpsz);
		fptrtop=(FPTR*)(areatbl+ATBLSZ);
	}
    setfmap();
	fpx=fpy=0;
	cp_x1=cp_y1=cp_x2=cp_y2=0;
}

/*** Set fmapmode ***/
static setfmap()
{
	if(fmapmode) {
		view_x=ebox_x;
		view_y=ebox_y;
		view_sx=view_sy=8;
		scrl_x=SCR_X1;
		scrl_y=SCR_Y1;
	} else {
		view_x=VBX_X;
		view_y=VBX_Y;
		view_sx=VBX_SX;
		view_sy=VBX_SY;
		scrl_x=SCR_X0;
		scrl_y=SCR_Y0;
		if(fsx-fpx < VBX_SX) fpx=fsx-VBX_SX;
		if(fsy-fpy < VBX_SY) fpy=fsy-VBX_SY;
	}
}

/*** Draw fmap ***/

drawfmap()
{
	drawmenu(&mb_fmap);
	shaderct(view_x,view_y,view_sx*32,view_sy*32,4);
	drawview();
	drawscrl(scrl_x,scrl_y,0,TRUE);
	if(fmapmode) {
		if(fullmap) drawmap(); else drawselw();
	}
}

/*** Do field map ***/

int do_fmap()
{
	flag f;
	
	f=FALSE;
	if(fmapmode) if(!(f=do_view())) f=do_selct();
	if(!f) f=do_scrl();
	return(f);
}

/*** Do view box ***/

static FMAP* setfmapp(x,y)
int x,y;
{
	return(fmaptop+y*fsx+x);
}

static int do_view()
{
	word x,y;
	word cx,cy;
	FMAP *mp;
	
	x=(msrec.px-view_x)/32;
	y=(msrec.py-view_y)/32;
	cx=x*32+view_x;
	cy=y*32+view_y;
	if(!inrect(x,y,8,8)) return(FALSE);
	mp=setfmapp(x+fpx,y+fpy);
	switch(fmapmode) {
	case MN_EDT:
		if(msrec.evnt & MS_LDN) {
			flushrct(cx,cy,32,32);
			selsprn=mp->spr;
			bgobst=mp->obst;
			drawobst();
		}
		else if(msrec.evnt & MS_RDN) {
			mp->spr=selsprn;
			mp->obst=bgobst;
			putfspr(x,y,selsprn,0);
		}
		break;
	case MN_CPY:
		if(msrec.evnt & MS_LDN) {
			flushrct(cx,cy,32,32);
			cp_x2=cp_x1;
			cp_y2=cp_y1;
			cp_x1=x+fpx;
			cp_y1=y+fpy;
		}
		else if(msrec.evnt & MS_RDN) {
			flushrct(cx,cy,32,32);
			if(dialog(&dg_sure)) {
				copyfmap(x+fpx,y+fpy);
				drawview();
			}
		}
		break;
	case MN_ARA:
		doarea(x,y);
		break;
	}
	return(TRUE);
}

/*** Copy map ***/

static copyfmap(cx,cy)
int  cx,cy;
{
	int  x,y;
	int  ax,ay;
	int  bx,by;
	int  dx,dy;
	FMAP *srcp,*dstp;
	
	if(cp_x1 <= cp_x2) { ax=cp_x1; bx=cp_x2; }
	else { ax=cp_x2; bx=cp_x1; }
	if(cp_y1 <= cp_y2) { ay=cp_y1; by=cp_y2; }
	else {	ay=cp_y2; by=cp_y1; }
	srcp=setfmapp(ax,ay);
	dstp=setfmapp(cx,cy);
	dx=cx-ax;
	dy=cy-ay;
	if(dstp < srcp) {
		for(y=ay; y<=by; y++) {
			for(x=ax; x<=bx; x++) {
				srcp=setfmapp(x,y);
				dstp=setfmapp(x+dx,y+dy);
				*(long*)dstp=*(long*)srcp;
			}
		}
	}
	else if(dstp > srcp) {
		for(y=by; y>=ay; y--) {
			for(x=bx; x>=ax; x--) {
				srcp=setfmapp(x,y);
				dstp=setfmapp(x+dx,y+dy);
				*(long*)dstp=*(long*)srcp;
			}
		}
	}
}

/*** Do scroll box ***/

static int do_scrl()
{
	int  fx,fy;
	int  dir;
	flag f;
	
	dir=doscrl(scrl_x,scrl_y);
	if(dir == -1) f=FALSE;
	else {
		fx=fpx;
		fy=fpy;
		if(dir > 0) {
			f=FALSE;
			if(dir & DR_RGT) { if(fpx < fsx-view_sx){ fx++; f=TRUE; }}
			else if(dir & DR_LFT) { if(fpx) 		{ fx--; f=TRUE; }}
			if(dir & DR_DWN) { if(fpy < fsy-view_sy){ fy++; f=TRUE; }}
			else if(dir & DR_UP)  { if(fpy) 		{ fy--; f=TRUE; }}
			if(f) {
				if(fullmap && fmapmode) {
					putscbx(0);
					fpx=fx;
					fpy=fy;
					putscbx(2);
				} else {
					fpx=fx;
					fpy=fy;
				}
				drawview();
			}
		}
		f=TRUE;
	}
	return(f);
}

/*** Save map ***/

savefmap()
{
	int  r;
	HEAD *hp;

	r=setname(fmapname,fmapmask+1);
	if(r == ERROR) goto end_save;
	if(r > 0) drawmenu(&mb_fmap);
	strcpy(fmapptr,fmapid);
	hp=(HEAD*)fmapptr;
	hp->size=fmapbyte;
	hp->fsx=fsx;
	hp->fsy=fsy;
	hp->areac=ATBLSZ;
	hp->fptrc=fptr_c;
	savefile(fmapname,fmapptr,fmapbyte);
end_save:
}

/*** Load map ***/

static loadfmap()
{
	char *name;
	HEAD *hp;
	
	if((name=(char*)selfile(fmapmask,NIL)) == NIL) goto end_load;
	freemem(fmapptr);
	fmapptr=(char*)loadfile(name);
	if(!fmapptr) goto err_load;
	if(strcmp(fmapptr,fmapid)) goto err_load;
	hp=(HEAD*)fmapptr;
	fmapbyte=hp->size;
	fsx=hp->fsx;
	fsy=hp->fsy;
	fptr_c=hp->fptrc;
	strcpy(fmapname,name);
	goto end_load;
err_load:	
	dialog(&dg_lodng);
end_load:
}

/*** Draw full map ***/

drawmap()
{
	int  x,y;
	
	mousoff();
	drawmenu(&mb_selw);
	pn.col=0;
	fillrect(sel_x,sel_y,256,256,NIL);
	map_s.pat=(char*)pt_map;
	map_s.col0=0;
	map_s.col1=7;
	for(y=sel_y; y<sel_y+fsy*4; y+=4) {
		for(x=sel_x; x<sel_x+fsx*4; x+=4) {
			fillrect(x,y,4,4,&map_s);
		}
	}
	putscbx(2);
	mouson();
}

/*** Draw Scroll box ***/

static putscbx(col)
int  col;
{
	pn.col=col;
	framerct(sel_x+fpx*4-1,sel_y+fpy*4-1,33,33);
}

/*** Do full map ***/

dofulmap()
{
	int  x,y;
	int  cx,cy;
	int  evnt;
	
	cx=msrec.px-sel_x-fpx*4;
	cy=msrec.py-sel_y-fpy*4;
	if(inrect(cx,cy,32,32) && msrec.evnt & MS_LDN) {
		repeat {
			if((evnt=getmous(&msrec)) & MS_MOV) {
				x=(int)(msrec.px-sel_x-cx)/4;
				y=(int)(msrec.py-sel_y-cy)/4;
				if(x < 0) x=0;
				else if(x > fsx-8) x=fsx-8;
				if(y < 0) y=0;
				else if(y > fsy-8) y=fsy-8;
				if(x != fpx || y != fpy) {
					mousoff();
					putscbx(0);
					fpx=x;
					fpy=y;
					putscbx(2);
					mouson();
				}
			}
		} until(evnt & MS_LUP);
		drawview();
	}
}

/*** Get New Area No. ***/

static int getarea()
{
	int  i;		
	int  n,m;
	FMAP *fp;
	
	fp=fmaptop;
	n=0;
	for_to(i,fsx*fsy) {
		m=fp->n;
		if(m > n && m < 128) n=m;
		fp++;
	}	
	return(n+1);
}

static int gettree(n1,n2)
int  n1,n2;
{
	int  i;
	TREE *p;
	
	p=areatbl;
	for_to(i,ATBLSZ) {
		if(p->n1 == n1 && p->n2 == n2) break;
		if(!((p++)->n1)) break; 
	}
	return(i+ATBLSZ);
}

/*** Do area ***/

static doarea(x,y)
int  x,y;
{
	int  n;
	int  m;
	FMAP *fp;
	TREE *tp;
	
	fp=setfmapp(x+fpx,y+fpy);
	if(msrec.evnt & MS_LDN+MS_RDN) {
		if(msrec.evnt & MS_RDN || fp->n) {
			selarea=fp->n;
			fmapp=fp;
			drawview();
		} else {
			n=getarea();
			dg_area.link=(CTRL*)&sb_count;
			sb_count.val=n;
			if(dialog(&dg_area) == Q_OK) {
				n=sb_count.val;
				fp->n=selarea=n;
				fmapp=fp;
				drawview();
			}
		}
	}
	else if(msrec.btn & MS_LON+MS_RON && fp != fmapp) {
		fmapp=fp;
		if(!fp->n) fp->n=selarea;
		else {
			n=fp->n;
			if(n != selarea) {
				if(msrec.btn & MS_LON) {
					if(n < selarea) m=selarea;
					else {
						m=n;
						n=selarea;
					}
					tp=chknp(fp->n=gettree(n,m));
					tp->n1=n;
					tp->n2=m;
				} else fp->n=selarea;
			}
		}
		putfspr(x,y,fp->spr,2);
	}
}

/*** Draw Area number ***/

drawarea(n)
int  n;
{
	int  x,y;
	int  prtp;
	
	x=ebox_x;
	y=ebox_y-18;
	prtp=pinit(PR_GRAF);
	mousoff();
	pn.col=7;
	fillrect(x,y,ANUMSX,16,NIL);
	shaderct(x,y,ANUMSX,16,1);
	gloc(x,y);
	pn.col=0;
	print("Area:%d",n);
	mouson();
	pinit(prtp);
}

