/*******************
	Music Player
*******************/

#include "std.h"

#define HEADSZ	16

typedef struct header {
	char id[4];
	int  size;
	int  count;
	char dummy[8];
} HEAD;

typedef struct mblkrec {
	int  size;
	int  part;
	int  nall;
	char barlin;
	char beat;
	char name[24];
} MBLK;

enum playcrec {
	CP_ON,
	CP_OFF,
	CP_STOP,
	CP_MUTE,
	CP_SEEK,
	CP_BLCK,
	CP_STT,
	CP_POS
};

static HEAD *vheadp;
static HEAD *mheadp;
static char fname[13];
static char extmsc[]=".msc";
static char extvce[]=".vce";

/*** Subroutines ***/

static MBLK* seek(n)
int  n;
{
	int  i;
	MBLK *mp;
	
	mp=(MBLK*)(mheadp+1);
	for_to(i,n) {
		mp=(MBLK*)((char*)mp+mp->size);
	}
	return(mp);
}

static MBLK* display(n)
int  n;
{
	MBLK *mp;
	mp=seek(n);
	print("%3d     ",n+1);
	print("%17s",mp->name);
	print("%3d\n",mp->nall);
	return(mp);
}	

/*** load file ***/

int  load(name,ext)
char *name;
char *ext;
{
	int  loadp;
	char *p;
	
	strcpy(fname,name);
	p=fname;
	while(*p) {
		if(*p == '.') break;
		p++;
	}
	if(!*p) strcpy(p,ext);
	loadp=loadfile(fname);
	if(!loadp) {
		print("*** Can't find \"%s\"\n",fname);
	} else if(loadp == -1) {
		print("*** Not enough memory\n");
		loadp=0;
	}
	return(loadp);
}

/*************
	'PLAY'
*************/

main(argc,argv)
int  argc;
char *argv[];
{
	int  i;
	int  mcount;
	int  mnumb;
	int  volume;
	int  mh;
	int  key;
	int  tick;
	int  mode;
	int  bar,bar1;
	MBLK *mp;
	flag playf;

	print("\n=== Music Player Version 0.81 ===\n");
	print("     < by ZOBplus C.MOS >\n\n");
	if(argc < 3) {
		print("*** Command syntax ...\n");
		print(" play voice[.vce] music[.msc]\n\n");
		print("*** Key Operation ...\n");
		print(" [CR] ........ Play/Pause\n");
		print(" [],[].... Quick Search\n");
		print(" [DEL]........ Skip to Next Music\n");
		print(" [INS]........ Replay/Skip to Pre Music\n");
		print(" [],[] ... Up/Down Volume \n");
		print(" [ESC] ....... Escape from Player\n\n");
		goto error;
	}
	if(!(vheadp=(HEAD*)load(argv[1],extvce))) goto error;
	if(!(mheadp=(HEAD*)load(argv[2],extmsc))) goto error;
	initsnd();
	initvoic(vheadp);
	volume=15;
	setmvolm(volume);
	print("*** %s ***\n",argv[2]);
	print("<No>    <title>         <Bars>\n");
	mcount=mheadp->count;
	for_to(i,mcount) {
		display(i);
	}
	print("------------------------------\n");
	mnumb=0;
	playf=FALSE;
	mode=3;
	bar=-1;
	repeat {
		if(mode >= 3) {
			mp=(MBLK*)display(mnumb);
			mh=playmusc(mp,1);
		}
		if(mode == 4) {
			bar=mp->nall-1;
			if(playf) playctrl(CP_STOP,mh);
			playctrl(CP_SEEK,mh,bar-1);
			if(playf) playctrl(CP_ON,mh);
			bar=-1;
		}
		if(mode >= 2) {
			if(playf) {
				playctrl(CP_ON,mh);
				print("PLAY  :");
			} else {
				playctrl(CP_STOP,mh);
				print("PAUSE :");
			}
			prpt(SPC,3);
		}
		repeat {
			if(!playctrl(CP_STT,mh)) goto skip_f;
			bar1=playctrl(CP_POS,mh,&tick);
			if(bar1 && bar1 != bar) {
				bar=bar1;
				prpt(0x08,3);
				print("%3d",bar);
			}
			mode=0;
			key=gchr(5);
			switch(key) {
			case 0x0D:
				playf=!playf;
				mode=2;
				break;
			case 0x3C00:
				bar++;
				mode=1;
				break;
			case 0x3B00:
				if(bar > 1) {
					bar--;
					mode=1;
				} else {
					if(!(mnumb--)) mnumb=mcount-1;
					mode=4;
				}
				break;
			case 0x3900:
	skip_f:		if(++mnumb == mcount) mnumb=0;
				mode=3;
				break;
			case 0x3800:
				if(bar <= 1) {
					if(!(mnumb--)) mnumb=mcount-1;
				}
				mode=3;
				break;
			case 0x3A00:
				if(volume < 15) setmvolm(++volume);
				break;
			case 0x3D00:
				if(volume > 0) setmvolm(--volume);
				break;
			case 0x1B:
				mode=5;
				break;
			}
		} until(mode);
		switch(mode) {
		case 1:
			if(playf) playctrl(CP_STOP,mh);
			playctrl(CP_SEEK,mh,bar-1);
			if(playf) playctrl(CP_ON,mh);
			break;
		case 2:
			pchr(0x0D);
			break;
		case 3:
		case 4:
			pchr(0x0B);
			pchr(0x0D);
			playctrl(CP_OFF,mh);
			break;
		case 5:
			playctrl(CP_OFF,mh);
			break;
		}
		bar=-1;
	} until(mode == 5);	
	initsnd();
error:;
}
